{-
Print stuff from elements.xml

Currently, the command line is

	$ printelts options...

in which the options (with number of arguments) are:
	-d 1	your FullRecall directory
	-l 1	print n most difficult (hardest) items
	-r 1	print items more difficult than n
	-c 1	print all items of the given category
	-t 0	uses tabs rather than spaces to set columns in simple output
	-s 0	show category column
	-f 1	font to use for HTML output
	-h 1	output HTML rather than simple standard output
*	-p 1	output PDF rather than simple standard output
*	-n 0	report on the numbers of items shown and total
*	-x 0	extended formatted statistics, always to stderr
*	-i 0	output id numbers only

Starred items are not yet implemented in the Haskell version.

If you specify -p or -h, the default (simple) output is not produced.

Horizontal alignment in the plain text output is difficult;  currently, the
script presumes that the font is mono-spaced, which is clearly wrong, but I
know of no other way of calculating column-alignment.  Option -t helps here.

The script just outputs questions and answers in UTF-8, formatted into columns.
There is the possibility of converting this into PDF but currently the software
does not work (for me).  Instructions for doing this:

1.  You need a BDF font;  convert a TTF using http://crl.nmsu.edu/~mleisher/ttf2bdf.html

	$ otf2bdf-LINUX-3.0/otf2bdf -r 800 -v ./.fonts/Code2000.ttf -o Code2000.bdf

2.  UTF-8 Text to PS convertor:  http://oldrus-ispell.sourceforge.net/txtbdf2ps.html, development version.

	$ printelts | txtbdf2ps-dev.pl --bdf=Code2000.bdf -UTF-8 -page=A4 -tab=30 -nowrap >sample.ps

The problem I experienced was that the txtbdf2ps-dev script outputs Postscript
that GhostScript cannot convert.  GS 8.14 is alright;  8.57 is not.

Suggestions and improvements are solicited, and should be sent to
michael at the domain name of landcroft dot co dot uk.
-}

import Data.List
import Data.Ord
import System
import System.IO
import ReadFR
import ReadFRRaw	-- make this ReadFRXML to experiment with `real' XML parsing
import Text.Regex
import Text.Printf
import Time
import Codec.Binary.UTF8.String


main :: IO ()
main = do
		args <- getArgs
		rightNow <- getClockTime
		readAndParse $ parseArgs args $ initialParms rightNow


readAndParse :: PrintParms -> IO ()
readAndParse parms =
	do
		items <- itemStream $ directory parms ++ "elements.xml"
		sendOutput (fileHTML parms) parms items


sendOutput :: (Maybe String) -> PrintParms -> [FRItem] -> IO ()
sendOutput Nothing parms items = putStr $ unlines $ textFRstuff parms items
sendOutput (Just htName) parms items = writeFile htName $ unlines $ htmlFRstuff parms items


-- each learnt element is reduced to this:
data FRItemWgrade = FRItemWgrade { parts :: [String], grade :: Float, category :: [String] }


textFRstuff :: PrintParms -> [FRItem] -> [String]
textFRstuff parms itemList =
	("FullRecall items" ++ cateShow (reqdCategory parms)) :
	(concat $
		map (showFRitems (breakLine parms) (spacing parms) (justWidths (columns parms))) $
			listFRstuff parms itemList)


htmlFRstuff :: PrintParms -> [FRItem] -> [String]
htmlFRstuff parms itemList =
	"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">" :
	("<html><font face=\"" ++ (fontName parms) ++ "\">") :
	"<head>" :
	"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\"/>" :
	("<title>FullRecall Items" ++ (cateShow (reqdCategory parms)) ++ "</title></head>") :
	"<table width=\"100%\" cellpadding=\"2\" cellspacing=\"0\" border=\"1\">" :
	("<tr bgcolor=\"#ffcc00\">" ++ (concat $ map colHTMLhead (columns parms)) ++ "</tr>") :
	(concat $ map tableRow $ listFRstuff parms itemList) ++
	[ "</table>" ] ++
{-
		-- if $show_numbers then
		--	html.print "<br>Items: ", @results.length, "/", @items, " (", 100.0 * @results.length / @items, "%)\n"
		-- end
-}
	[ "</body>", "</html>" ]
	where
		colHTMLhead (name, _) = "<th>" ++ name ++ "</th>"
		tableRow item = "<tr>" : (concat $ map htPart (parts item)) : [ "</tr>" ]
		htPart string = "<td>" ++ (regCData $ regImage $ regBreak $ regCData $ regImage string) ++ "</td>"


listFRstuff :: PrintParms -> [FRItem] -> [FRItemWgrade]
listFRstuff parms itemList =
	chopTop (topSoMany parms) $
		sortBy gradeFRcompare $
			filter (filterFR parms) $
				map (reformFRitem parms) itemList


gradeFRcompare :: FRItemWgrade -> FRItemWgrade -> Ordering
gradeFRcompare one two =
	if grade one > grade two then
		LT
	else
		if grade one < grade two then GT else EQ


cateShow :: [String] -> String
cateShow [] = []
cateShow repr = " (" ++ (concat $ intersperse "/" repr) ++ ")"


chopTop :: (Maybe Int) -> ([a] -> [a])
chopTop Nothing = id
chopTop (Just howMany) = take howMany


reformFRitem :: PrintParms -> FRItem -> FRItemWgrade
reformFRitem parms item =
	FRItemWgrade columnValues thisGrade (ReadFR.category item)
	where
		columnValues = onlyRequired $ map reqdCol [(catName, categoryName item), (qName, question item), (aName, answer item), (gradeName, printf "%7.2f" thisGrade) ]
		reqdCol = extractReqdColumns (columns parms)
		thisGrade = grading (timeStamp parms) item


filterFR :: PrintParms -> FRItemWgrade -> Bool
filterFR parms item = correctCategory && scoreFloor
	where
		correctCategory = isPrefixOf (reqdCategory parms) (reverse $ Main.category item)
		scoreFloor = grade item >= harderThan parms


breakLine :: PrintParms -> String
breakLine parms = replicate (sum $ justWidths $ columns parms) '='


justWidths :: ([ColumnSpec] -> [Int])
justWidths = map (\(_, value) -> value)


showFRitems :: String -> Columns -> [Int] -> FRItemWgrade -> [String]
showFRitems divideLine colSplitting colWidths oneFRitem = 
	divideLine : (columnsIntoRows colSplitting outputRowByRow)
	where
		-- the columns, split to fit width
		outputRowByRow :: [ColumnData]
		outputRowByRow = map makePart $ zip colWidths $ parts oneFRitem

		makePart :: ColumnPair -> ColumnData
		makePart (colWidth, colValue) = (colWidth, splitByWidth colWidth $ regCData $ regImage $ regFinal $ regInitial colValue)


-- are we outputting this column ?
extractReqdColumns :: [ColumnSpec] -> (String, String) -> (Maybe ColumnPair)
extractReqdColumns columnSpecs (name, value) = colReqd value $ lookup name columnSpecs


colReqd :: String -> (Maybe Int) -> (Maybe ColumnPair)
colReqd _ Nothing = Nothing
colReqd name (Just width) = Just (width, name)


onlyRequired :: [Maybe ColumnPair] -> [String]
onlyRequired [] = []
onlyRequired (Nothing:xs) = onlyRequired xs
onlyRequired (Just (_, str) : xs) = str : onlyRequired xs


type ColumnSpec = (String, Int)	-- one column in the textual output

type ColumnPair = (Int, String)	-- one string for output, with its width

type ColumnData = (Int, [String])	-- one string for output, split to fit


columnsIntoRows :: Columns -> [ColumnData] -> [String]
columnsIntoRows colHow widthStringPairs =
	if allNull widthStringPairs then
		[]
	else
		(padRow colHow (map softHead widthStringPairs)) : (columnsIntoRows colHow (map softTail widthStringPairs))


allNull :: ([(a, [b])] -> Bool)
allNull = all (\(_, oneStr) -> null oneStr)


softHead :: ColumnData -> ColumnPair
softHead (width, []) = (width, "")
softHead (width, (x:_)) = (width, x)


softTail :: ColumnData -> ColumnData
softTail (width, []) = (width, [])
softTail (width, (_:xs)) = (width, xs)


padRow :: Columns -> [ColumnPair] -> String
padRow _ [] = error "insufficient pads"
padRow colHow ((width, string):xs) =
	if allNull xs then
		string
	else
		string ++ interSpacing colHow ++ padOne xs
	where
		interSpacing Tabbing = "\t"
		interSpacing Spaces = replicate (width - length string) ' '


padOne :: [ColumnPair] -> String
padOne [] = error "insufficient pads"
padOne ((width, string):xs) =
	if allNull xs then
		string
	else
		string ++ (replicate (width - length string) ' ') ++ padOne xs


regInitial, regFinal, regImage, regCData, regBreak :: String -> String
regInitial string = subRegex (mkRegex "<[bus]>") string "{"
regFinal string = subRegex (mkRegex "</[bus]>") string "}"
regImage string = subRegex (mkRegex "<img>.*</img>") string ""
regCData string = subRegex (mkRegexWithOpts "^<\\!\\[CDATA\\[(.*)\\]\\]>$" False True) string "\\1"
regBreak string = subRegex (mkRegexWithOpts "\n" False True) string "<br/>"


qName, aName, catName, gradeName :: String
catName = "category"
gradeName = "gr."
qName = "Q"
aName = "A"


data Pair a = Pair { first, second :: a }


splitByWidth :: Int -> String -> [String]
splitByWidth n item = map encodeString $ splitByWidthEnc n $ decodeString item


splitByWidthEnc :: Int -> String -> [String]
-- Given a width and a string, return the string broken up into segments that
-- will fit into the width.
splitByWidthEnc _ [] = []
splitByWidthEnc n item =
	(take (first breakpoint) item) : splitByWidthEnc n (drop (second breakpoint) item)
	where
		breakpoint = findBreak (Pair n n) n 0 item


findBreak :: (Pair Int) -> Int -> Int -> String -> Pair Int
-- Return the point at which to break a string to fit it into a fixed width of
-- column.  It tries to fit as much as possible.  Since the break might very
-- well be at blank space, we return two values:  an `up to' & a `resume from'.
findBreak _ _ at [] = Pair at at
findBreak _ _ at ('\n':_) = Pair at (at+1)
findBreak _ limit at (' ':xs) = findBreak (Pair at (at + nextPoint + 1)) limit (at+nextPoint+1) (drop nextPoint xs)
	where
		nextPoint = firstNonSpace xs

findBreak bestSoFar limit at (_:xs) = if at >= limit then bestSoFar else findBreak bestSoFar limit (at+1) xs


firstNonSpace :: String -> Int
firstNonSpace (' ':xs) = 1 + firstNonSpace xs
firstNonSpace _ = 0


-- How do we split columns in the plain text output ?
-- with spaces, or just a single tab ?
data Columns = Spaces | Tabbing


-- The parameters which can vary the behaviour of the module;  these can be
-- controlled by command-line parameters;  see function parseArgs.
data PrintParms = PrintParms {
	useXML :: Bool,
	fontName :: String,
	spacing :: Columns,
	directory :: String,
	fileHTML :: Maybe String,
	reqdCategory :: [String],
	timeStamp :: Int,
	harderThan :: Float,
	topSoMany :: Maybe Int,
	columns :: [ColumnSpec] }

initialParms :: ClockTime -> PrintParms
initialParms now = PrintParms False "Code2000" Spaces "./" Nothing [] (timeAsUTC now) 0.0 Nothing [ (qName, 35), (aName, 44) ]


secsInDay :: Int
secsInDay = 24 * 60 * 60

secsInYear :: Int
secsInYear = secsInDay * 365	-- I know, I know ...

timeAsUTC :: ClockTime -> Int
timeAsUTC time =
	(ctYear timeUTC - 1970) * secsInYear + ctYDay timeUTC * secsInDay
	where
		timeUTC = toUTCTime time

-- various arguments:
parseArgs :: [String] -> PrintParms -> PrintParms

parseArgs ("-f":newfont:rest) parms = parseArgs rest parms {fontName = newfont}

parseArgs ("-l":soMany:rest) parms = parseArgs rest parms {topSoMany = Just (read soMany) }

parseArgs ("-r":minimal:rest) parms = parseArgs rest parms {harderThan = read minimal, columns = (columns parms) ++ [(gradeName, 7)] }

parseArgs ("-d":dirName:rest) parms = parseArgs rest parms {directory = regularise dirName "/"}

parseArgs ("-x":rest) parms = parseArgs rest parms {useXML = True}

parseArgs ("-t":rest) parms = parseArgs rest parms {spacing = Tabbing}

parseArgs ("-c":justThisCate:rest) parms = parseArgs rest parms {reqdCategory = breakAt '/' justThisCate}

parseArgs ("-s":rest) parms = parseArgs rest parms {columns = (catName, 20) : (columns parms)}

parseArgs ("-h":htmlname:rest) parms = parseArgs rest parms {fileHTML = (Just (regularise htmlname ".html"))}

parseArgs (what:_) _ = error $ '\'' : what ++ "': invalid argument"

parseArgs [] parms = parms


breakAt :: Char -> String -> [String]
breakAt _ [] = []
breakAt c list = breakParts $ span (/= c) list
	where
		breakParts (x, theRest) = x : breakRest theRest
			where
				breakRest [] = []
				breakRest (_:xs) = breakAt c xs


grading :: Int -> FRItem -> Float
grading startStamp item = log $ 1.0 + metric_luohan
	where
		attrs = attributes item
		metric_luohan = sqrt $ intervalRatio * repeatRatio
		intervalRatio = repetitions * repetitions / (interval + 1.0)
		interval = findFloatValue "ivl" attrs
		repetitions = findFloatValue "rp" attrs
		repeatRatio = repetitions / ((lifeTime / 86400) ** 0.38)
		lifeTime = fromIntegral startStamp - (findFloatValue "id" attrs)

findFloatValue :: String -> [(String, String)] -> Float
findFloatValue name list = convertFloat $ lookup name list

convertFloat :: (Maybe String) -> Float
convertFloat Nothing = error "attribute not found"
convertFloat (Just str) = read str


-- regularise ensures that file names have the required ending,
-- by appending it if they don't.
regularise :: String -> String -> String
regularise string ending =
	let
		str_len = length string
		end_len = length ending
	in
		if str_len < end_len ||
				(drop (str_len - end_len) string /= ending) then
			string ++ ending
		else
			string
